from pathlib import Path

from alembic import command, config

from poaster.core import sessions


def upgrade_to_head() -> None:
    """Upgrade database to head from within the app."""
    alembic_cfg_path = Path(__file__).parent / "alembic.ini"
    alembic_cfg = config.Config(alembic_cfg_path)

    with sessions.engine.begin() as connection:
        alembic_cfg.attributes["connection"] = connection
        command.upgrade(alembic_cfg, "head")



from typing import Iterable, TypedDict

import haitch as H
import mistletoe as md

from . import config


def page(content: H.SupportsHtml, *, title: str, username: str = "") -> H.Element:
    """Base HTML container for building web pages."""
    return H.html5(
        content=H.fragment(
            navbar(username),
            content,
            footer(),
        ),
        page_title=f"{title} - {config.settings.title}",
        language_code="en",
        body_classes="flex-wrapper",
        links=(
            H.link(rel="stylesheet", href="/static/normalize.css"),
            H.link(rel="stylesheet", href="/static/concrete.css"),
            H.link(rel="stylesheet", href="/static/app.css"),
        ),
    )


def not_found_page(*, username: str) -> H.Element:
    """Not found 404 page."""
    content = message("Page not found.", variant="warning")
    return page(content, title="404 - Not found", username=username)


def unauthorized_page() -> H.Element:
    """Unauthorized 401 page."""
    content = message("Unauthorized: please login.", variant="danger")
    return page(content, title="401 - Unauthorized")


class LinkProp(TypedDict):
    """Anchor link prop."""

    name: str
    href: str
    show: bool


def links(props: Iterable[LinkProp]) -> H.Element:
    """List of URL links."""
    style = "padding-right: 1.7rem;"
    return H.span(
        H.a(href=prop["href"], style=style)(prop["name"])
        for prop in props
        if prop["show"]
    )


def logo() -> H.Element:
    """Logo component that links to home."""
    logo_style = (
        "padding-right: 3rem;"
        "text-decoration: none;"
        f"color: {config.settings.get_color(variant='primary')};"
        "font-family: monospace;"
    )
    return H.a(href="/", style=logo_style)(config.settings.title)


def navbar(username: str) -> H.Element:
    """Navbar component."""
    nav_style = "flex-direction: row; padding: 2rem 0;"
    nav_links: Iterable[LinkProp] = [
        {"name": "+add", "href": "/posts/new", "show": username != ""},
        {"name": "logout", "href": "/logout", "show": username != ""},
        {"name": "login", "href": "/login", "show": not username},
    ]
    return H.header(
        H.nav(style=nav_style)(
            logo(),
            links(nav_links),
        )
    )


def footer() -> H.Element:
    """Footer component."""
    return H.footer(
        H.small("poaster "),
        H.a(href="https://sr.ht/~loges/poaster/")(H.small("[source]")),
    )


def message(text: str, variant: config.ColorVariant) -> H.Element:
    """Message component for with consistent styling based on variant provided."""
    color = config.settings.get_color(variant)
    style = (
        "padding: 2rem;"
        f"border-left: thick solid {color};"
        f"background: {color}13;"
        "padding-left: 1.2em;"
    )
    return H.p(style=style)(text)


def markdown_to_html(text: str) -> H.Element:
    """Render markdown text to HTML."""
    with md.HtmlRenderer() as renderer:
        rendered_html = renderer.render(md.Document(text))
    return H.unsafe(rendered_html)



import click
import uvicorn
import uvicorn.config

from poaster.__about__ import __version__
from poaster.access.cli import users
from poaster.core import config, hashing
from poaster.migrations.upgrade import upgrade_to_head


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(version=__version__, prog_name="poaster")
def poaster() -> None:
    """Control panel for managing poaster application."""


@click.command()
def init() -> None:
    """Instantiate the application environment and secret key."""
    click.secho("Secret key for application:", fg="green")

    if config.settings.secret_key:
        click.echo("Key already found in your environment: `SECRET_KEY`\n")
    else:
        click.echo(f"SECRET_KEY={hashing.generate_secret_token()}")
        click.secho("Copy and paste this into your `.env` file.\n", fg="yellow")

    click.secho("Migrating database to head:", fg="green")
    upgrade_to_head()
    click.echo("Successfully migrated to head.")


@click.command()
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Bind socket to this host.",
    show_default=True,
)
@click.option(
    "--port",
    type=int,
    default=8000,
    help="Bind socket to this port.",
    show_default=True,
)
def run(host: str, port: int) -> None:
    """Migrate database to latest version and launch application server."""

    click.secho("Migrating database to head:", fg="green")
    upgrade_to_head()
    click.echo("Successfully migrated to head.\n")

    click.secho("Starting server...", fg="green")
    uvicorn.run(
        "poaster.app:app",
        host=host,
        port=port,
        log_level=config.settings.log_level,
    )


poaster.add_command(init)
poaster.add_command(run)
poaster.add_command(users)



from fastapi import HTTPException, Request, Response
from fastapi.responses import HTMLResponse
from starlette import status

from . import components

InvalidCredentials = HTTPException(
    status_code=status.HTTP_401_UNAUTHORIZED,
    detail="Could not validate credentials",
    headers={"WWW-Authenticate": "Bearer"},
)

Forbidden = HTTPException(
    status_code=status.HTTP_403_FORBIDDEN,
    detail="Action not authorized",
)

NotFound = HTTPException(
    status_code=status.HTTP_404_NOT_FOUND,
    detail="Resource not found.",
)


async def handle_forbidden(request: Request, exc: Exception) -> Response:
    return HTMLResponse(
        str(components.unauthorized_page()),
        status_code=status.HTTP_403_FORBIDDEN,
    )


async def handle_not_found(request: Request, exc: Exception) -> Response:
    return HTMLResponse(
        str(components.not_found_page(username="")),
        status_code=status.HTTP_404_NOT_FOUND,
    )


async def handle_validation_error(request: Request, exc: Exception) -> Response:
    return HTMLResponse(
        str(components.not_found_page(username="")),
        status_code=status.HTTP_404_NOT_FOUND,
    )



import secrets

from passlib.context import CryptContext

from . import config

pwd_context = CryptContext(schemes=["argon2"])


def generate_secret_token() -> str:
    """Helper for generating a secret key."""
    return secrets.token_hex(config.settings.secret_key_n_bytes)



from sqlalchemy import event
from sqlalchemy.engine import create_engine
from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine

from poaster.core.config import settings

engine = create_engine(settings.uri)
async_engine = create_async_engine(settings.async_uri)
async_session = async_sessionmaker(async_engine, expire_on_commit=False)


@event.listens_for(async_engine.sync_engine, "connect")
def set_sqlite_pragma(dbapi_connection, _) -> None:  # type: ignore
    """Set sqlite settings before instantiating a connection."""
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()



import haitch as H

from poaster.bulletin import schemas, services
from poaster.core import components


def post_form_new() -> H.Element:
    """Form for adding a new post."""
    return H.form(id_="newPost", action="/posts/new", method="post")(
        H.h1("Add new post"),
        H.label(for_="title")("Title:"),
        H.input(id_="title", type_="text", name="title"),
        H.label(for_="text")("Text:"),
        H.textarea(id_="text", name="text", placeholder="Markdown supported", rows=8),
        H.br(),
        H.a(href="/posts", style="padding-right: 1.5rem;")("Back"),
        H.button(type_="submit")("Save"),
    )


def post_form_edit(post: schemas.PostSchema) -> H.Element:
    """Form for editing a post."""
    return H.form(id_="editPost", action=f"/posts/{post.id}/edit", method="post")(
        H.h1("Edit post"),
        H.label(for_="title")("Title:"),
        H.input(id_="title", type_="text", name="title", value=post.title),
        H.label(for_="text")("Text:"),
        H.textarea(id_="text", name="text", rows=8)(post.text),
        H.br(),
        components.links([{"name": "Back", "href": f"/posts/{post.id}", "show": True}]),
        H.button(type_="submit")("Save"),
    )


def post_form_delete(post: schemas.PostSchema) -> H.Element:
    """Form for editing a post."""
    return H.form(id_="deletePost", action=f"/posts/{post.id}/delete", method="post")(
        H.h1("Delete post"),
        components.message(
            "Are you sure you want to permanently delete this post?", variant="danger"
        ),
        H.br(),
        components.links([{"name": "Back", "href": f"/posts/{post.id}", "show": True}]),
        H.button(type_="submit")("Confirm"),
    )


async def post_list(bulletin_service: services.BulletinService) -> H.Child:
    """List of stored posts."""
    if not (posts := await bulletin_service.get_posts()):
        return components.message("No posts were found.", variant="warning")

    return [
        H.section(
            H.h2(post.title),
            await post_meta_info(bulletin_service, post),
            H.p(H.a(href=f"/posts/{post.id}")("View")),
        )
        for post in posts
    ]


async def post_detail(
    bulletin_service: services.BulletinService, post: schemas.PostSchema, username: str
) -> H.Element:
    """Detail regarding a post."""
    return H.fragment(
        H.h1(post.title),
        await post_meta_info(bulletin_service, post),
        components.markdown_to_html(post.text),
        H.br(),
        components.links(
            [
                {
                    "name": "Back",
                    "href": "/posts",
                    "show": True,
                },
                {
                    "name": "Edit",
                    "href": f"/posts/{post.id}/edit",
                    "show": username != "",
                },
                {
                    "name": "Delete",
                    "href": f"/posts/{post.id}/delete",
                    "show": username != "",
                },
            ]
        ),
    )


async def post_meta_info(
    bulletin_service: services.BulletinService, post: schemas.PostSchema
) -> H.Element:
    """Meta information regarding a post."""
    latest_post = await bulletin_service.get_latest_version_of_post(id=post.id)
    updated_at_str = latest_post.updated_at.strftime("%b %d, %Y at %H:%M")
    return H.p(H.small(f"Last updated on {updated_at_str} UTC"))



import asyncio
import logging

import poaster.access.repository
import poaster.access.schemas
import poaster.access.services
import poaster.bulletin.repository
import poaster.bulletin.schemas
import poaster.bulletin.services
from poaster.core import exceptions, sessions


async def load_fixtures():
    """Load test fixtures useful for local development."""
    async with sessions.async_session() as session:
        access_svc = poaster.access.services.AccessService.from_session(session)
        bulletin_svc = poaster.bulletin.services.BulletinService.from_session(session)

        await add_dummy_user(access_svc)
        await add_dummy_posts(bulletin_svc)


async def add_dummy_user(access_svc: poaster.access.services.AccessService):
    try:
        await access_svc.register_user(username="dummy", password="password")
    except exceptions.AlreadyExists:
        logging.info("'dummy' user already exists with password equal to 'password'.")
    else:
        logging.info("Added 'dummy' user with password equal to 'password'.")


async def add_dummy_posts(bulletin_svc: poaster.bulletin.services.BulletinService):
    async def add_post(title: str, text: str):
        await bulletin_svc.create_post(
            username="dummy",
            payload=poaster.bulletin.schemas.PostInputSchema(title=title, text=text),
        )

    await add_post("Penguins", "Penguins are a group of aquatic flightless birds.")
    await add_post("Tigers", "Tigers are the largest living cat species.")
    await add_post("Koalas", "Koala is is native to Australia.")

    logging.info("Added example dummy posts about animals.")


if __name__ == "__main__":
    asyncio.run(load_fixtures())



"""Add post versions table.

Revision ID: 34b9b6b1c9a3
Revises: 821b409e627c
Create Date: 2024-03-02 17:19:54.189806

"""
import asyncio
import logging

import sqlalchemy as sa
from alembic import op

from poaster.bulletin import repository, schemas
from poaster.core import exceptions, sessions

# revision identifiers, used by Alembic.
revision = "34b9b6b1c9a3"
down_revision = "821b409e627c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "post_versions",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("post_id", sa.Integer(), nullable=True),
        sa.Column("title", sa.String(length=255), nullable=False),
        sa.Column("text", sa.Text(), nullable=False),
        sa.Column("version", sa.Integer(), nullable=False),
        sa.Column("updated_by", sa.String(length=255), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(["post_id"], ["posts.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("post_id", "version", name="unique_post_version"),
    )
    asyncio.run(initialize_post_versions())


def downgrade() -> None:
    op.drop_table("post_versions")


async def initialize_post_versions() -> None:
    """Initialize existing posts with with version 1."""
    async with sessions.async_session() as session:
        post_repo = repository.SqlalchemyPostRepository(session)
        post_version_repo = repository.SqlalchemyPostVersionRepository(session)

        for post in await post_repo.get_all():
            try:
                await post_version_repo.get_by_id(post_id=post.id, version_id=1)
            except exceptions.DoesNotExist:
                logging.info("'%s' initialized with version 1.", post.title)
                await post_version_repo.create(
                    username=post.created_by,
                    post_id=post.id,
                    post=schemas.PostInputSchema(title=post.title, text=post.text),
                )
            else:
                logging.debug("'%s' post already initialized.", post.title)



from __future__ import annotations

import dataclasses
from typing import Optional

from sqlalchemy.ext.asyncio import AsyncSession

from poaster.bulletin import repository, schemas
from poaster.core import exceptions


@dataclasses.dataclass(frozen=True)
class BulletinService:
    """The service layer for the bulletin domain."""

    _post_repo: repository.SupportsPostRepository
    _post_version_repo: repository.SupportsPostVersionRepository

    @classmethod
    def from_session(cls, session: AsyncSession) -> BulletinService:
        """Build bulletin service from database session."""
        return cls(
            _post_repo=repository.SqlalchemyPostRepository(session),
            _post_version_repo=repository.SqlalchemyPostVersionRepository(session),
        )

    async def create_post(
        self, *, username: str, payload: schemas.PostInputSchema
    ) -> schemas.PostSchema:
        """Create a bulletin post from an authenticated user."""
        post = await self._post_repo.create(username, payload)

        await self._post_version_repo.create(
            username=username,
            post_id=post.id,
            post=payload,
        )

        return post

    async def update_post(
        self, *, id: int, username: str, payload: schemas.PostInputSchema
    ) -> Optional[schemas.PostSchema]:
        """Get a bulletin post by id."""
        try:
            post = await self._post_repo.update(id, payload)
        except exceptions.DoesNotExist:
            return None

        await self._post_version_repo.create(
            username=username,
            post_id=post.id,
            post=payload,
        )

        return post

    async def delete_post(self, *, id: int) -> Optional[schemas.PostSchema]:
        """Get a bulletin post by id."""
        try:
            post = await self._post_repo.delete(id)
        except exceptions.DoesNotExist:
            return None
        else:
            return post

    async def get_post(self, *, id: int) -> Optional[schemas.PostSchema]:
        """Get a bulletin post by id."""
        try:
            return await self._post_repo.get_by_id(id)
        except exceptions.DoesNotExist:
            return None

    async def get_posts(self) -> list[schemas.PostSchema]:
        """Get all recipes."""
        return await self._post_repo.get_all()

    async def get_latest_version_of_post(self, *, id: int) -> schemas.PostVersionSchema:
        """Get the latest version of a bulletin post by id."""
        return await self._post_version_repo.get_latest(post_id=id)



"""Add users table.

Revision ID: 0b3ee652b174
Revises: 
Create Date: 2024-02-09 22:05:49.244176

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0b3ee652b174"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "users",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("username", sa.String(length=255), nullable=False),
        sa.Column("password", sa.String(length=512), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("username"),
    )


def downgrade() -> None:
    op.drop_table("users")



import asyncio

import click
import pydantic

from poaster.core import exceptions, sessions

from . import services


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
def users() -> None:
    """Control panel for managing users."""


@click.command("add")
@click.option(
    "--username",
    type=str,
    prompt="Username",
    help="Username input. [prompt]",
)
@click.option(
    "--password",
    type=str,
    prompt="Password",
    hide_input=True,
    confirmation_prompt=True,
    help="Password input. [prompt]",
)
def add_user(username: str, password: str) -> None:
    """Add new user."""
    try:
        asyncio.run(add_user_(username, password))
    except exceptions.AlreadyExists:
        click.secho("User already exists.", fg="yellow")
    except pydantic.ValidationError as err:
        click.secho(f"Input validation failed: {err}", fg="yellow")
    else:
        click.secho(f"`{username}` successfully added.", fg="green")


async def add_user_(username: str, password: str) -> None:
    async with sessions.async_session() as session:
        access_service = services.AccessService.from_session(session)
        await access_service.register_user(username, password)
        await session.commit()


@click.command("list")
def list_usernames() -> None:
    """List stored usernames."""
    if usernames := asyncio.run(list_usernames_()):
        click.secho("Stored users:", fg="green")
        for username in sorted(usernames):
            click.echo(f"- {username}")
    else:
        click.secho("No users found.", fg="yellow")


async def list_usernames_() -> list[str]:
    async with sessions.async_session() as session:
        access_service = services.AccessService.from_session(session)
        return await access_service.list_usernames()


users.add_command(add_user)
users.add_command(list_usernames)



from collections.abc import AsyncGenerator

import pytest
from click.testing import CliRunner
from fastapi.testclient import TestClient
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker
from sqlalchemy.ext.asyncio.session import AsyncSession


@pytest.fixture(scope="session", autouse=True)
async def db_engine() -> AsyncGenerator[AsyncEngine, None]:
    from poaster.core import sessions, tables

    async with sessions.async_engine.begin() as conn:
        await conn.run_sync(tables.Base.metadata.create_all)

    yield sessions.async_engine

    async with sessions.async_engine.begin() as conn:
        await conn.run_sync(tables.Base.metadata.drop_all)


@pytest.fixture
async def db_session(db_engine: AsyncEngine) -> AsyncGenerator[AsyncSession, None]:
    """Database session that will rolled back after test run."""
    from poaster.core import tables

    async with async_sessionmaker(db_engine, expire_on_commit=False)() as session:
        yield session
        await session.rollback()

        for table in reversed(tables.Base.metadata.sorted_tables):
            await session.execute(text(f"DELETE FROM {table.name};"))
            await session.commit()


@pytest.fixture(scope="session")
def client() -> TestClient:
    """Test client (httpx) fixture for making requests to the app."""
    from poaster.app import app

    return TestClient(app)


@pytest.fixture(scope="session")
def cli_runner() -> CliRunner:
    return CliRunner()



import datetime

import pytest
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession

from poaster.bulletin import repository, schemas
from poaster.core import exceptions


@pytest.fixture
def post_repo(db_session: AsyncSession) -> repository.SqlalchemyPostRepository:
    return repository.SqlalchemyPostRepository(db_session)


@pytest.fixture
def post_version_repo(
    db_session: AsyncSession,
) -> repository.SqlalchemyPostVersionRepository:
    return repository.SqlalchemyPostVersionRepository(db_session)


async def test_create_post(post_repo: repository.SqlalchemyPostRepository):
    got = await post_repo.create(
        username="testuser",
        post=schemas.PostInputSchema(title="hi", text="hello, world!"),
    )
    want = schemas.PostSchema(
        id=1,
        title="hi",
        text="hello, world!",
        created_by="testuser",
        created_at=got.created_at,
    )

    assert got == want


async def test_update_post(
    db_session: AsyncSession, post_repo: repository.SupportsPostRepository
):
    qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(qry))

    got = await post_repo.update(
        1, schemas.PostInputSchema(title="hi v2", text="hello, world!!")
    )
    want = schemas.PostSchema(
        id=got.id,
        title="hi v2",
        text="hello, world!!",
        created_by="testuser",
        created_at=datetime.datetime(2024, 2, 13, 8, 0),
    )

    assert got == want


async def test_delete_post_response(
    db_session: AsyncSession, post_repo: repository.SupportsPostRepository
):
    qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(qry))

    got = await post_repo.delete(1)
    want = schemas.PostSchema(
        id=got.id,
        title="hi",
        text="hello, world!",
        created_by="testuser",
        created_at=datetime.datetime(2024, 2, 13, 8, 0),
    )

    assert got == want


async def test_delete_post_actually_deleted(
    db_session: AsyncSession, post_repo: repository.SupportsPostRepository
):
    qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(qry))
    await post_repo.delete(1)
    with pytest.raises(exceptions.DoesNotExist, match="Post doesn't exist."):
        await post_repo.get_by_id(1)


async def test_delete_post_cascade_deletes_versions(
    db_session: AsyncSession,
    post_repo: repository.SupportsPostRepository,
    post_version_repo: repository.SupportsPostVersionRepository,
):
    add_post_qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(add_post_qry))

    add_post_version_qry = """
    INSERT INTO post_versions (post_id, title, text, version, updated_by, updated_at)
    VALUES (1, 'hi', 'hello, world!', 1, 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(add_post_version_qry))

    await post_repo.delete(1)

    with pytest.raises(exceptions.DoesNotExist, match="Post version doesn't exist."):
        await post_version_repo.get_by_id(post_id=1, version_id=1)


async def test_get_post_by_id(
    db_session: AsyncSession, post_repo: repository.SupportsPostRepository
):
    qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(qry))

    got = await post_repo.get_by_id(1)
    want = schemas.PostSchema(
        id=got.id,
        title="hi",
        text="hello, world!",
        created_by="testuser",
        created_at=datetime.datetime(2024, 2, 13, 8, 0),
    )

    assert got == want


async def test_post_not_found(post_repo: repository.SupportsPostRepository):
    with pytest.raises(exceptions.DoesNotExist, match="Post doesn't exist."):
        await post_repo.get_by_id(42)


async def test_get_all_posts(
    db_session: AsyncSession, post_repo: repository.SupportsPostRepository
):
    qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES
        ('oldest_post', 'hello, oldest_post!', 'testuser', '2024-02-13 08:00'),
        ('newest_post', 'hello, newest_post!', 'testuser', '2024-02-15 12:00')
    ;
    """
    await db_session.execute(text(qry))

    got = await post_repo.get_all()
    want = [
        schemas.PostSchema(
            id=2,
            title="newest_post",
            text="hello, newest_post!",
            created_by="testuser",
            created_at=datetime.datetime(2024, 2, 15, 12, 0),
        ),
        schemas.PostSchema(
            id=1,
            title="oldest_post",
            text="hello, oldest_post!",
            created_by="testuser",
            created_at=datetime.datetime(2024, 2, 13, 8, 0),
        ),
    ]

    assert got == want


async def test_get_all_none_found(post_repo: repository.SupportsPostRepository):
    got = await post_repo.get_all()
    want = []

    assert got == want


async def test_create_post_version_v1(
    db_session: AsyncSession,
    post_version_repo: repository.SqlalchemyPostVersionRepository,
):
    add_post_qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(add_post_qry))

    got = await post_version_repo.create(
        username="testuser",
        post_id=1,
        post=schemas.PostInputSchema(title="hi", text="hello, world!"),
    )
    want = schemas.PostVersionSchema(
        id=1,
        post_id=1,
        title="hi",
        text="hello, world!",
        version=1,
        updated_by="testuser",
        updated_at=got.updated_at,
    )

    assert got == want


async def test_create_post_version_v2(
    db_session: AsyncSession,
    post_version_repo: repository.SqlalchemyPostVersionRepository,
):
    add_post_qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(add_post_qry))

    add_post_version_qry = """
    INSERT INTO post_versions (post_id, title, text, version, updated_by, updated_at)
    VALUES (1, 'hi', 'hello, world!', 1, 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(add_post_version_qry))

    got = await post_version_repo.create(
        username="testuser",
        post_id=1,
        post=schemas.PostInputSchema(title="hi", text="hello, world!"),
    )
    want = schemas.PostVersionSchema(
        id=2,
        post_id=1,
        title="hi",
        text="hello, world!",
        version=2,
        updated_by="testuser",
        updated_at=got.updated_at,
    )

    assert got == want


async def test_get_all_post_versions(
    db_session: AsyncSession,
    post_version_repo: repository.SqlalchemyPostVersionRepository,
):
    add_post_qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(add_post_qry))

    add_post_versions_qry = """
    INSERT INTO post_versions (post_id, title, text, version, updated_by, updated_at)
    VALUES
      (1, 'hi v1', 'hello, world!', 1, 'testuser', '2024-02-13 08:00'),
      (1, 'hi v2', 'hello, world!!', 2, 'testuser', '2024-02-13 09:00'),
      (1, 'hi v3', 'hello, world!!!', 3, 'testuser', '2024-02-13 10:00')
    ;
    """
    await db_session.execute(text(add_post_versions_qry))

    got = await post_version_repo.get_all(post_id=1)
    want = [
        schemas.PostVersionSchema(
            id=1,
            post_id=1,
            title="hi v1",
            text="hello, world!",
            version=1,
            updated_by="testuser",
            updated_at=datetime.datetime(2024, 2, 13, 8),
        ),
        schemas.PostVersionSchema(
            id=2,
            post_id=1,
            title="hi v2",
            text="hello, world!!",
            version=2,
            updated_by="testuser",
            updated_at=datetime.datetime(2024, 2, 13, 9),
        ),
        schemas.PostVersionSchema(
            id=3,
            post_id=1,
            title="hi v3",
            text="hello, world!!!",
            version=3,
            updated_by="testuser",
            updated_at=datetime.datetime(2024, 2, 13, 10),
        ),
    ]

    assert got == want


async def test_get_latest_version(
    db_session: AsyncSession,
    post_version_repo: repository.SqlalchemyPostVersionRepository,
):
    add_post_qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(add_post_qry))

    add_post_versions_qry = """
    INSERT INTO post_versions (post_id, title, text, version, updated_by, updated_at)
    VALUES
      (1, 'hi v1', 'hello, world!', 1, 'testuser', '2024-02-13 08:00'),
      (1, 'hi v2', 'hello, world!!', 2, 'testuser', '2024-02-13 09:00'),
      (1, 'hi v3', 'hello, world!!!', 3, 'testuser', '2024-02-13 10:00')
    ;
    """
    await db_session.execute(text(add_post_versions_qry))

    got = await post_version_repo.get_latest(post_id=1)
    want = schemas.PostVersionSchema(
        id=3,
        post_id=1,
        title="hi v3",
        text="hello, world!!!",
        version=3,
        updated_by="testuser",
        updated_at=datetime.datetime(2024, 2, 13, 10),
    )

    assert got == want


async def test_get_post_version_by_id(
    db_session: AsyncSession,
    post_version_repo: repository.SqlalchemyPostVersionRepository,
):
    add_post_qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(add_post_qry))

    add_post_versions_qry = """
    INSERT INTO post_versions (post_id, title, text, version, updated_by, updated_at)
    VALUES
      (1, 'hi v1', 'hello, world!', 1, 'testuser', '2024-02-13 08:00'),
      (1, 'hi v2', 'hello, world!!', 2, 'testuser', '2024-02-13 09:00'),
      (1, 'hi v3', 'hello, world!!!', 3, 'testuser', '2024-02-13 10:00')
    ;
    """
    await db_session.execute(text(add_post_versions_qry))

    got = await post_version_repo.get_by_id(post_id=1, version_id=2)
    want = schemas.PostVersionSchema(
        id=2,
        post_id=1,
        title="hi v2",
        text="hello, world!!",
        version=2,
        updated_by="testuser",
        updated_at=datetime.datetime(2024, 2, 13, 9),
    )

    assert got == want



from sqlalchemy.sql import func
from sqlalchemy.sql.schema import Column
from sqlalchemy.sql.sqltypes import DateTime, Integer, String

from poaster.core.tables import Base


class User(Base):
    __tablename__ = "users"

    id = Column(Integer, primary_key=True)
    username = Column(String(255), nullable=False, unique=True)
    password = Column(String(512), nullable=False)
    created_at = Column(DateTime, default=func.now())



import pytest
from fastapi.testclient import TestClient
from sqlalchemy.ext.asyncio import AsyncSession
from starlette import status

from poaster.access import repository, schemas
from poaster.access.services import AccessService


@pytest.fixture(autouse=True)
async def add_testuser(db_session: AsyncSession):
    repo = repository.SqlalchemyUserRepository(db_session)
    user = schemas.UserRegistrationSchema(username="testuser", password="secret")
    await repo.create(user)


def test_generate_token_success(client: TestClient):
    payload = {"username": "testuser", "password": "secret"}
    response = client.post("/api/auth/token", data=payload)
    assert response.status_code == status.HTTP_200_OK


def test_generate_token_fails_with_bad_username(client: TestClient):
    payload = {"username": "notvalid", "password": "secret"}
    response = client.post("/api/auth/token", data=payload)
    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert response.json() == {"detail": "Could not validate credentials"}


def test_generate_token_fails_with_bad_password(client: TestClient):
    payload = {"username": "testuser", "password": "notvalid"}
    response = client.post("/api/auth/token", data=payload)
    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert response.json() == {"detail": "Could not validate credentials"}


@pytest.mark.parametrize(
    "payload",
    [
        pytest.param({}, id="empty payload"),
        pytest.param({"username": "testuser"}, id="only username"),
        pytest.param({"password": "secret"}, id="only password"),
        pytest.param({"foo": "bar"}, id="invalid fields"),
    ],
)
def test_generate_token_fails_with_invalid_payload(
    client: TestClient, payload: dict[str, str]
):
    response = client.post("/api/auth/token", data=payload)
    assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY


def test_fetch_current_user_success(client: TestClient):
    token = AccessService.create_access_token(username="testuser")
    headers = {"authorization": f"Bearer {token}"}
    response = client.get("/api/auth/me", headers=headers)
    assert response.status_code == status.HTTP_200_OK
    assert response.json() == {"username": "testuser"}


def test_fetch_current_user_without_token_header_fails(client: TestClient):
    response = client.get("/api/auth/me", headers={})
    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert response.json() == {"detail": "Not authenticated"}


def test_fetch_current_user_with_bad_header_fails(client: TestClient):
    headers = {"authorization": "Bearer blahblahblahblah"}
    response = client.get("/api/auth/me", headers=headers)
    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert response.json() == {"detail": "Could not validate credentials"}



import haitch as H
from fastapi import APIRouter, Form, Response, status
from fastapi.responses import HTMLResponse, RedirectResponse

from poaster import dependencies
from poaster.core import components

from . import components as local_components

router = APIRouter()


@router.get("/login")
async def handle_display_login_page(
    username: dependencies.UsernameFromSessionCookie,
) -> Response:
    """Defines web endpoint for displaying the login page."""
    if username:
        return RedirectResponse("/posts", status_code=status.HTTP_302_FOUND)

    content = H.main(local_components.login_form())

    page = components.page(content, title="Login", username=username)

    return HTMLResponse(str(page))


@router.post("/login")
async def handle_login(
    access_service: dependencies.AccessService,
    username: str = Form(...),
    password: str = Form(...),
) -> Response:
    """Defines web endpoint for authenticating user credentials."""
    user = await access_service.authenticate(username, password)

    if user is None:
        return RedirectResponse("/login/failed", status_code=status.HTTP_302_FOUND)

    response = RedirectResponse("/posts", status_code=status.HTTP_302_FOUND)
    response.set_cookie("session", access_service.create_access_token(user.username))

    return response


@router.get("/logout")
async def handle_logout() -> Response:
    """Defines web endpoint for authenticating user credentials."""
    response = RedirectResponse("/login", status_code=status.HTTP_302_FOUND)
    response.delete_cookie("session")
    return response


@router.get("/login/failed")
async def handle_display_login_failed() -> HTMLResponse:
    """Defines web endpoint for displaying the login failure."""
    content = H.main(
        components.message("Failed to login.", variant="danger"),
        local_components.login_form(),
    )

    page = components.page(content, title="Login Failed")

    return HTMLResponse(str(page))



from sqlalchemy.sql import func
from sqlalchemy.sql.schema import Column, ForeignKey, UniqueConstraint
from sqlalchemy.sql.sqltypes import DateTime, Integer, String, Text

from poaster.core.tables import Base


class Post(Base):
    __tablename__ = "posts"

    id = Column(Integer, primary_key=True)
    title = Column(String(255), nullable=False)
    text = Column(Text(), nullable=False)
    created_by = Column(String(255), nullable=False)
    created_at = Column(DateTime, default=func.now())


class PostVersion(Base):
    __tablename__ = "post_versions"

    id = Column(Integer, primary_key=True)
    title = Column(String(255), nullable=False)
    text = Column(Text(), nullable=False)
    version = Column(Integer, nullable=False)
    updated_by = Column(String(255), nullable=False)
    updated_at = Column(DateTime, default=func.now())

    post_id = Column(Integer, ForeignKey("posts.id", ondelete="CASCADE"))

    __table_args__ = (
        UniqueConstraint("post_id", "version", name="unique_post_version"),
    )



from typing import Any

import pytest
from fastapi.testclient import TestClient
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession
from starlette import status

from poaster.access.services import AccessService

Headers = dict[str, str]
"""Helper alias for request header (only for testing)."""

Payload = dict[str, Any]
"""Helper alias for request payload (only for testing)."""


@pytest.fixture
async def add_example_posts(db_session: AsyncSession) -> None:
    qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES
        ('oldest_post', 'hello, oldest_post!', 'testuser', '2024-02-13 08:00'),
        ('newest_post', 'hello, newest_post!', 'testuser', '2024-02-15 12:00')
    ;
    """
    await db_session.execute(text(qry))


@pytest.fixture
def headers() -> Headers:
    token = AccessService.create_access_token(username="testuser")
    return {"authorization": f"Bearer {token}"}


def test_create_post(client: TestClient, headers: Headers):
    payload = {"title": "hi", "text": "hello, world!"}
    response = client.post("/api/posts", headers=headers, json=payload)
    assert response.status_code == status.HTTP_201_CREATED


def test_create_post_unauthorized(client: TestClient):
    payload = {"title": "hi", "text": "hello, world!"}
    response = client.post("/api/posts", data=payload)
    assert response.status_code == status.HTTP_401_UNAUTHORIZED


@pytest.mark.parametrize(
    "payload",
    [
        pytest.param({}, id="empty payload"),
        pytest.param({"title": "hi"}, id="only title"),
        pytest.param({"text": "hello, world!"}, id="only text"),
        pytest.param({"title": 1, "text": "hello, world!"}, id="bad title"),
        pytest.param({"title": "hi", "text": 1}, id="bad text"),
        pytest.param({"title": "hi" * 1000, "text": "hello"}, id="too long of a title"),
    ],
)
def test_create_post_with_bad_payload(
    client: TestClient, headers: Headers, payload: Payload
):
    response = client.post("/api/posts", headers=headers, json=payload)
    assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY


@pytest.mark.usefixtures("add_example_posts")
def test_update_post(client: TestClient, headers: Headers):
    payload = {"title": "hi v2", "text": "hello, world!!"}
    response = client.put("/api/posts/1", headers=headers, json=payload)
    assert response.status_code == status.HTTP_200_OK


def test_update_post_not_found(client: TestClient, headers: Headers):
    payload = {"title": "hi v2", "text": "hello, world!!"}
    response = client.put("/api/posts/42", headers=headers, json=payload)
    assert response.status_code == status.HTTP_404_NOT_FOUND


def test_update_post_unauthorized(client: TestClient):
    payload = {"title": "hi", "text": "hello, world!"}
    response = client.put("/api/posts/1", data=payload)
    assert response.status_code == status.HTTP_401_UNAUTHORIZED


@pytest.mark.parametrize(
    "payload",
    [
        pytest.param({}, id="empty payload"),
        pytest.param({"title": "hi"}, id="only title"),
        pytest.param({"text": "hello, world!"}, id="only text"),
        pytest.param({"title": 1, "text": "hello, world!"}, id="bad title"),
        pytest.param({"title": "hi", "text": 1}, id="bad text"),
        pytest.param({"title": "hi" * 1000, "text": "hello"}, id="too long of a title"),
    ],
)
def test_update_post_with_bad_payload(
    client: TestClient, headers: Headers, payload: Payload
):
    response = client.put("/api/posts/1", headers=headers, json=payload)
    assert response.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY


@pytest.mark.usefixtures("add_example_posts")
def test_delete_post(client: TestClient, headers: Headers):
    response = client.delete("/api/posts/1", headers=headers)
    assert response.status_code == status.HTTP_200_OK


def test_delete_post_not_found(client: TestClient, headers: Headers):
    response = client.delete("/api/posts/42", headers=headers)
    assert response.status_code == status.HTTP_404_NOT_FOUND


def test_delete_post_unauthorized(client: TestClient):
    response = client.delete("/api/posts/1")
    assert response.status_code == status.HTTP_401_UNAUTHORIZED


@pytest.mark.usefixtures("add_example_posts")
def test_get_post(client: TestClient):
    response = client.get("/api/posts/1")
    assert response.status_code == status.HTTP_200_OK


def test_get_post_not_found(client: TestClient):
    response = client.get("/api/posts/42")
    assert response.status_code == status.HTTP_404_NOT_FOUND


@pytest.mark.usefixtures("add_example_posts")
def test_get_posts(client: TestClient):
    response = client.get("/api/posts")
    assert response.status_code == status.HTTP_200_OK



from fastapi import APIRouter
from starlette.status import HTTP_201_CREATED

from poaster import dependencies
from poaster.core import http_exceptions

from . import schemas

router = APIRouter(tags=["bulletin"])


@router.post(
    "/posts",
    status_code=HTTP_201_CREATED,
    summary="Creates post when passed valid input.",
)
async def handle_create_post(
    payload: schemas.PostInputSchema,
    username: dependencies.UsernameFromBearerToken,
    bulletin_service: dependencies.BulletinService,
) -> schemas.PostSchema:
    """Defines endpoint for creating bulletin posts."""
    return await bulletin_service.create_post(username=username, payload=payload)


@router.put("/posts/{id}", summary="Update post by id.")
async def handle_update_post(
    id: int,
    payload: schemas.PostInputSchema,
    username: dependencies.UsernameFromBearerToken,
    bulletin_service: dependencies.BulletinService,
) -> schemas.PostSchema:
    """Defines endpoint for updating a post."""
    post = await bulletin_service.update_post(id=id, username=username, payload=payload)

    if post is None:
        raise http_exceptions.NotFound

    return post


@router.delete("/posts/{id}", summary="Delete post by id.")
async def handle_delete_post(
    id: int,
    _: dependencies.UsernameFromBearerToken,
    bulletin_service: dependencies.BulletinService,
) -> schemas.PostSchema:
    """Defines endpoint for deleting a post."""
    if (post := await bulletin_service.delete_post(id=id)) is None:
        raise http_exceptions.NotFound

    return post


@router.get("/posts", summary="Get all posts.")
async def handle_get_posts(
    bulletin_service: dependencies.BulletinService,
) -> list[schemas.PostSchema]:
    """Defines endpoint for retrieving all posts."""
    return await bulletin_service.get_posts()


@router.get("/posts/{id}", summary="Get post by id.")
async def handle_get_post(
    id: int,
    bulletin_service: dependencies.BulletinService,
) -> schemas.PostSchema:
    """Defines endpoint for retrieving a post by its id."""
    if (post := await bulletin_service.get_post(id=id)) is None:
        raise http_exceptions.NotFound

    return post



from __future__ import annotations

import dataclasses
import datetime
from typing import Optional

from sqlalchemy.ext.asyncio import AsyncSession

from poaster.core import exceptions, hashing, oauth

from . import repository, schemas


@dataclasses.dataclass(frozen=True)
class AccessService:
    """The service layer for the access domain."""

    _user_repo: repository.SupportsUserRepository

    @classmethod
    def from_session(cls, session: AsyncSession) -> AccessService:
        """Build access service from database session."""
        return cls(_user_repo=repository.SqlalchemyUserRepository(session))

    async def authenticate(
        self, username: str, password: str
    ) -> Optional[schemas.UserSchema]:
        """Check that input credentials match the user's stored credentials."""
        try:
            user = await self._user_repo.get_by_username(username)
        except exceptions.DoesNotExist:
            return None

        if not hashing.pwd_context.verify(password, user.password):
            return None

        return user

    async def check_username_exists(self, username: str) -> bool:
        """Check if username exists."""
        try:
            await self._user_repo.get_by_username(username)
        except exceptions.DoesNotExist:
            return False
        else:
            return True

    async def register_user(self, username: str, password: str) -> schemas.UserSchema:
        """Register user given valid username and password."""
        user = schemas.UserRegistrationSchema(username=username, password=password)
        return await self._user_repo.create(user)

    async def list_usernames(self) -> list[str]:
        """List out all persisted usernames."""
        return [user.username for user in await self._user_repo.get_all()]

    @classmethod
    def create_access_token(cls, username: str, minutes: int = 60) -> oauth.Token:
        """Generate user access token for a duration of time."""
        payload: oauth.TokenPayload = {
            "sub": username,
            "exp": datetime.datetime.utcnow() + datetime.timedelta(minutes=minutes),
        }
        return oauth.encode_token(payload)

    @classmethod
    def decode_access_token(cls, token: oauth.Token) -> schemas.UserTokenPayload:
        """Decode user access token."""
        payload = oauth.decode_token(token)
        return schemas.UserTokenPayload.model_validate(payload)



import haitch as H


def login_form() -> H.Element:
    """Login form for authenticating users."""
    return H.fragment(
        H.h1("Login"),
        H.form(action="/login", method="post")(
            H.fieldset(
                H.legend("Log in to add or edit posts"),
                H.label(for_="username")("Username:"),
                H.input(type_="text", name="username", required=True),
                H.label(for_="password")("Password:"),
                H.input(type_="password", name="password", required=True),
                H.button(type_="submit")("Login"),
            ),
        ),
    )



import datetime

import pydantic
import pytest
from sqlalchemy.ext.asyncio import AsyncSession

from poaster.access import schemas, services
from poaster.core import exceptions, oauth


@pytest.fixture
def access_service(db_session: AsyncSession) -> services.AccessService:
    return services.AccessService.from_session(db_session)


@pytest.fixture
async def user(access_service: services.AccessService) -> schemas.UserSchema:
    user = schemas.UserRegistrationSchema(username="bob", password="password")
    return await access_service._user_repo.create(user)


async def test_authenticate_success(
    access_service: services.AccessService, user: schemas.UserSchema
):
    got = await access_service.authenticate(user.username, "password")
    want = user

    assert got == want


async def test_authenticate_bad_username(access_service: services.AccessService):
    got = await access_service.authenticate("badusername", "password")
    want = None

    assert got == want


async def test_authenticate_bad_pw(
    access_service: services.AccessService, user: schemas.UserSchema
):
    got = await access_service.authenticate(user.username, "badpw")
    want = None

    assert got == want


def test_encoding_and_decoding_of_user_access_token(
    access_service: services.AccessService, user: schemas.UserSchema
):
    now_in_secs = int(datetime.datetime.now().strftime("%s"))

    token = access_service.create_access_token(username=user.username, minutes=1)
    payload = access_service.decode_access_token(token)

    assert payload.sub == user.username  # subject is username
    assert payload.exp - now_in_secs == 60  # expiration is 1 minute


def test_wrongly_formatted_token_raises_unauthorized(
    access_service: services.AccessService,
):
    token = oauth.Token("blahblahblahblahIamatokenhaha")

    with pytest.raises(exceptions.Unauthorized):
        access_service.decode_access_token(token)


def test_wrong_payload_field_raises_validation_error(
    access_service: services.AccessService,
):
    token = oauth.encode_token(
        {
            "sub": "bob",
            "iss": "me",  # 'iss' field is not part of the access token schema
        }
    )

    with pytest.raises(pydantic.ValidationError):
        access_service.decode_access_token(token)


@pytest.mark.parametrize(
    "creds",
    [
        pytest.param(("testuser", "password" * 100), id="password too long"),
        pytest.param(("testuser" * 100, "password"), id="username too long"),
    ],
)
async def test_register_user_validation(
    access_service: services.AccessService, creds: tuple[str, str]
):
    username, password = creds
    with pytest.raises(pydantic.ValidationError):
        await access_service.register_user(username, password)


async def test_register_user_and_list_usernames(access_service: services.AccessService):
    await access_service.register_user("testuser", "password")

    got = await access_service.list_usernames()
    want = ["testuser"]

    assert got == want


async def test_register_user_already_exists(access_service: services.AccessService):
    await access_service.register_user("testuser", "password")

    got = await access_service.check_username_exists("testuser")
    want = True

    assert got == want


async def test_check_username_does_not_exists(access_service: services.AccessService):
    await access_service.register_user("testuser", "password")

    got = await access_service.check_username_exists("baduser")
    want = False

    assert got == want


async def test_duplicate_user_registration(access_service: services.AccessService):
    await access_service.register_user("testuser", "password")
    with pytest.raises(exceptions.AlreadyExists):
        await access_service.register_user("testuser", "password")



from typing import Protocol

import sqlalchemy.exc
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select

from poaster.core import exceptions, hashing

from . import schemas, tables


class SupportsUserRepository(Protocol):
    """Interface for handling users."""

    async def create(self, user: schemas.UserRegistrationSchema) -> schemas.UserSchema:
        """Create user after validating input schema."""
        ...

    async def get_all(self) -> list[schemas.UserSchema]:
        """Fetch all users from DB."""
        ...

    async def get_by_username(self, username: str) -> schemas.UserSchema:
        """Fetch by username, raising exception if not found."""
        ...


class SqlalchemyUserRepository:
    """Implementation of the user repository with SqlAlchemy."""

    def __init__(self, session: AsyncSession) -> None:
        self._session = session

    async def create(self, user: schemas.UserRegistrationSchema) -> schemas.UserSchema:
        entry = tables.User(
            **user.model_dump(exclude={"password"}),
            password=hashing.pwd_context.hash(user.password),
        )
        self._session.add(entry)
        try:
            await self._session.commit()
            return schemas.UserSchema.model_validate(entry)
        except sqlalchemy.exc.IntegrityError as err:
            await self._session.rollback()
            raise exceptions.AlreadyExists(err) from err

    async def get_all(self) -> list[schemas.UserSchema]:
        results = await self._session.execute(select(tables.User))
        return [schemas.UserSchema.model_validate(res) for res in results.scalars()]

    async def get_by_username(self, username: str) -> schemas.UserSchema:
        qry = select(tables.User).where(tables.User.username == username).limit(1)
        result = await self._session.execute(qry)

        try:
            user = result.scalars().one()
        except sqlalchemy.exc.NoResultFound as err:
            raise exceptions.DoesNotExist(f"{username!r} does not exist.") from err

        return schemas.UserSchema.model_validate(user)



from typing import Any, NewType

from fastapi.security import OAuth2PasswordBearer
from jose import JWTError, jwt

from . import config, exceptions

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/api/auth/token")


Token = NewType("Token", str)
"""Represents an encoded JWT string."""

TokenPayload = dict[str, Any]
"""Alias for token payload encoded in JWT."""


def encode_token(payload: TokenPayload) -> Token:
    """Encode token payload using secret key and algorithm."""
    jwt_token = jwt.encode(
        payload,
        key=config.settings.secret_key,
        algorithm=config.settings.algorithm,
    )
    return Token(jwt_token)


def decode_token(token: Token) -> TokenPayload:
    """Decode token using secret key and algorithm."""
    try:
        return jwt.decode(
            token=token,
            key=config.settings.secret_key,
            algorithms=[config.settings.algorithm],
        )
    except JWTError as err:
        raise exceptions.Unauthorized from err



from fastapi import APIRouter

from poaster import dependencies
from poaster.core import http_exceptions

from . import schemas

router = APIRouter(prefix="/auth", tags=["access"])


@router.post("/token", summary="Generates access token when passed valid credentials.")
async def handle_generate_access_token(
    form: dependencies.AuthForm,
    access_service: dependencies.AccessService,
) -> schemas.AccessToken:
    """Defines endpoint for generating access tokens."""
    user = await access_service.authenticate(form.username, form.password)

    if user is None:
        raise http_exceptions.InvalidCredentials

    token = access_service.create_access_token(user.username)

    return schemas.AccessToken(access_token=token, token_type="bearer")


@router.get("/me", summary="Fetches information regarding current user.")
async def handle_get_current_user(
    username: dependencies.UsernameFromBearerToken,
    access_service: dependencies.AccessService,
) -> schemas.UserPublicSchema:
    """Defines endpoint for fetching information for the current user."""
    if await access_service.check_username_exists(username):
        return schemas.UserPublicSchema(username=username)

    raise http_exceptions.InvalidCredentials



import datetime

from pydantic import Field

from poaster.core.schemas import BaseSchema


class PostInputSchema(BaseSchema):
    """Post input from user."""

    title: str = Field(max_length=255)
    text: str


class PostSchema(BaseSchema):
    """Post fetched from the persistence layer."""

    id: int
    title: str
    text: str
    created_by: str
    created_at: datetime.datetime


class PostVersionSchema(BaseSchema):
    """Post version fetched from the persistence layer."""

    id: int
    post_id: int
    title: str
    text: str
    version: int
    updated_by: str
    updated_at: datetime.datetime



"""Add posts table.

Revision ID: 821b409e627c
Revises: 0b3ee652b174
Create Date: 2024-02-12 21:09:39.972955

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "821b409e627c"
down_revision = "0b3ee652b174"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "posts",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("title", sa.String(length=255), nullable=False),
        sa.Column("text", sa.Text(), nullable=False),
        sa.Column("created_by", sa.String(length=255), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade() -> None:
    op.drop_table("posts")



import pytest
from click.testing import CliRunner

from poaster import cli


@pytest.fixture
def patch_alembic_migration(monkeypatch: pytest.MonkeyPatch):
    def patch():
        print("patched migrations...")

    monkeypatch.setattr(cli, "upgrade_to_head", patch)


@pytest.fixture
def patch_generate_secret_token(monkeypatch: pytest.MonkeyPatch):
    def patch() -> str:
        return "PATCHED-KEY-123ABC"

    monkeypatch.setattr(cli.hashing, "generate_secret_token", patch)


@pytest.fixture
def patch_uvicorn_server(monkeypatch: pytest.MonkeyPatch):
    def patch(*args, **kwargs):
        print("patched server received the following params:")
        print(f"args: {args}")
        print(f"kwargs: {kwargs}")

    monkeypatch.setattr(cli.uvicorn, "run", patch)


def test_calling_poaster_without_args_launches_help(cli_runner: CliRunner):
    expected_help_text = """\
Usage: poaster [OPTIONS] COMMAND [ARGS]...

  Control panel for managing poaster application.

Options:
  --version   Show the version and exit.
  -h, --help  Show this message and exit.

Commands:
  init   Instantiate the application environment and secret key.
  run    Migrate database to latest version and launch application server.
  users  Control panel for managing users.
"""
    result = cli_runner.invoke(cli.poaster, [])

    assert result.exit_code == 0
    assert result.output == expected_help_text


def test_init_help(cli_runner: CliRunner):
    expected_help_text = """\
Usage: init [OPTIONS]

  Instantiate the application environment and secret key.

Options:
  --help  Show this message and exit.
"""
    result = cli_runner.invoke(cli.init, ["--help"])

    assert result.exit_code == 0
    assert result.output == expected_help_text


@pytest.mark.usefixtures("patch_alembic_migration")
def test_init_with_secret_key(cli_runner: CliRunner):
    cli_runner.env = {"SECRET_KEY": "ALREADY-HAVE-A-KEY-SET"}
    result = cli_runner.invoke(cli.init, [])
    expected_output = """\
Secret key for application:
Key already found in your environment: `SECRET_KEY`

Migrating database to head:
patched migrations...
Successfully migrated to head.
"""

    assert result.exit_code == 0
    assert result.output == expected_output


@pytest.mark.skip(reason="Need to find a way to patch local dotenv file.")
@pytest.mark.usefixtures("patch_alembic_migration", "patch_generate_secret_token")
def test_init_without_secret_key(cli_runner: CliRunner):
    cli_runner.env = {"SECRET_KEY": None}
    result = cli_runner.invoke(cli.init, [])
    expected_output = """\
Secret key for application:
SECRET_KEY=PATCHED-KEY-123ABC
Copy and paste this into your `.env` file.

Migrating database to head:
patched migrations...
Successfully migrated to head.
"""

    assert result.exit_code == 0
    assert result.output == expected_output


def test_run_help(cli_runner: CliRunner):
    expected_help_text = """\
Usage: run [OPTIONS]

  Migrate database to latest version and launch application server.

Options:
  --host TEXT     Bind socket to this host.  [default: 127.0.0.1]
  --port INTEGER  Bind socket to this port.  [default: 8000]
  --help          Show this message and exit.
"""
    result = cli_runner.invoke(cli.run, ["--help"])

    assert result.exit_code == 0
    assert result.output == expected_help_text


@pytest.mark.usefixtures("patch_alembic_migration", "patch_uvicorn_server")
def test_run_with_defaults(cli_runner: CliRunner):
    result = cli_runner.invoke(cli.run, [])
    expected_output = """\
Migrating database to head:
patched migrations...
Successfully migrated to head.

Starting server...
patched server received the following params:
args: ('poaster.app:app',)
kwargs: {'host': '127.0.0.1', 'port': 8000, 'log_level': 'info'}
"""

    assert result.exit_code == 0
    assert result.output == expected_output



from pydantic import Field

from poaster.core.schemas import BaseSchema


class UserLoginSchema(BaseSchema):
    """Login credentials for a user."""

    username: str = Field(max_length=255)
    password: str = Field(max_length=255)


class UserRegistrationSchema(BaseSchema):
    """Login credentials used for registration."""

    username: str = Field(max_length=255)
    password: str = Field(max_length=512)


class UserSchema(BaseSchema):
    """User fetched from the persistence layer."""

    id: int
    username: str
    password: str


class UserPublicSchema(BaseSchema):
    """User info to return to public api (no sensitive info)."""

    username: str


class UserTokenPayload(BaseSchema):
    """Payload for user access token."""

    sub: str
    exp: int


class AccessToken(BaseSchema):
    """Generated access token payload (JWT)."""

    access_token: str
    token_type: str



from collections.abc import AsyncGenerator

from fastapi import Depends
from fastapi.security import APIKeyCookie, OAuth2PasswordRequestForm
from sqlalchemy.ext.asyncio import AsyncSession
from typing_extensions import Annotated

import poaster.access.repository
import poaster.access.services
import poaster.bulletin.repository
import poaster.bulletin.services
from poaster.core import exceptions, http_exceptions, oauth, sessions

AuthBearer = Annotated[oauth.Token, Depends(oauth.oauth2_scheme)]
AuthForm = Annotated[OAuth2PasswordRequestForm, Depends()]
SessionCookie = Annotated[
    oauth.Token, Depends(APIKeyCookie(name="session", auto_error=False))
]


async def get_db_session() -> AsyncGenerator[AsyncSession, None]:
    """Get async database session, commiting changes before closing."""
    async with sessions.async_session() as session:
        yield session
        await session.commit()


DBSession = Annotated[AsyncSession, Depends(get_db_session)]


def get_access_service(db: DBSession) -> poaster.access.services.AccessService:
    """Instantiate user repository session."""
    return poaster.access.services.AccessService.from_session(db)


AccessService = Annotated[
    poaster.access.services.AccessService, Depends(get_access_service)
]


def get_bulletin_service(db: DBSession) -> poaster.bulletin.services.BulletinService:
    """Instantiate bulletin application service."""
    return poaster.bulletin.services.BulletinService.from_session(db)


BulletinService = Annotated[
    poaster.bulletin.services.BulletinService, Depends(get_bulletin_service)
]


def get_username_from_bearer_token(token: AuthBearer) -> str:
    """Try and retrieve username based on passed token."""
    try:
        payload = oauth.decode_token(token)
    except exceptions.Unauthorized as err:
        raise http_exceptions.InvalidCredentials from err
    else:
        return str(payload.get("sub", ""))


UsernameFromBearerToken = Annotated[str, Depends(get_username_from_bearer_token)]


def get_username_from_session_cookie(user_session: SessionCookie) -> str:
    """Try and retrieve username from session cookie."""
    if user_session is None:
        return ""
    try:
        payload = oauth.decode_token(user_session)
    except exceptions.Unauthorized:
        return ""
    else:
        return str(payload.get("sub", ""))


UsernameFromSessionCookie = Annotated[str, Depends(get_username_from_session_cookie)]



from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config, pool

from poaster.access.tables import User
from poaster.bulletin.tables import Post
from poaster.core.config import settings
from poaster.core.tables import Base

# Need to import all tables that should be autodetected by migration tool.
__all__ = [
    "Base",
    "Post",
    "User",
]

# User defined options
config = context.config
config.set_main_option("sqlalchemy.url", settings.uri)
target_metadata = Base.metadata

assert config.config_file_name is not None, "Cannot load file config without file name."

fileConfig(config.config_file_name)


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    configuration = config.get_section(config.config_ini_section)

    assert configuration is not None, "Cannot create engine without config"

    connectable = engine_from_config(
        configuration,
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(connection=connection, target_metadata=target_metadata)

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import haitch as H
from fastapi import APIRouter, Form, Response, status
from fastapi.responses import HTMLResponse, RedirectResponse

from poaster import dependencies
from poaster.bulletin import schemas
from poaster.core import components, http_exceptions

from . import components as local_components

router = APIRouter()


@router.get("/")
async def handle_homepage() -> Response:
    """Redirect user to posts page when the call root path."""
    return RedirectResponse("/posts", status_code=status.HTTP_302_FOUND)


@router.get("/posts")
async def handle_post_list(
    username: dependencies.UsernameFromSessionCookie,
    bulletin_service: dependencies.BulletinService,
) -> Response:
    """Defines web endpoint for displaying the login page."""
    content = H.main(
        await local_components.post_list(bulletin_service),
    )

    page = components.page(content, title="Post List", username=username)

    return HTMLResponse(str(page))


@router.get("/posts/new")
async def handle_post_new_page(
    username: dependencies.UsernameFromSessionCookie,
) -> Response:
    """Defines web endpoint for displaying new post to add."""
    if not username:
        raise http_exceptions.Forbidden

    content = H.main(local_components.post_form_new())

    page = components.page(content, title="Post New", username=username)

    return HTMLResponse(str(page))


@router.post("/posts/new")
async def handle_post_new(
    username: dependencies.UsernameFromSessionCookie,
    bulletin_service: dependencies.BulletinService,
    title: str = Form(...),
    text: str = Form(...),
) -> Response:
    """Defines web endpoint for handling new post."""
    if not username:
        raise http_exceptions.Forbidden

    await bulletin_service.create_post(
        username=username,
        payload=schemas.PostInputSchema(title=title, text=text),
    )

    return RedirectResponse("/posts", status_code=status.HTTP_302_FOUND)


@router.get("/posts/{id}")
async def handle_post_detail(
    id: int,
    username: dependencies.UsernameFromSessionCookie,
    bulletin_service: dependencies.BulletinService,
) -> Response:
    """Defines web endpoint for displaying post detail."""
    if (post := await bulletin_service.get_post(id=id)) is None:
        raise http_exceptions.NotFound

    content = H.main(
        await local_components.post_detail(bulletin_service, post, username)
    )

    page = components.page(content, title="Post Detail", username=username)

    return HTMLResponse(str(page))


@router.get("/posts/{id}/edit")
async def handle_post_edit_page(
    id: int,
    username: dependencies.UsernameFromSessionCookie,
    bulletin_service: dependencies.BulletinService,
) -> Response:
    """Defines web endpoint for displaying post to edit."""
    if not username:
        raise http_exceptions.Forbidden

    if (post := await bulletin_service.get_post(id=id)) is None:
        raise http_exceptions.NotFound

    content = H.main(local_components.post_form_edit(post))

    page = components.page(content, title="Post Edit", username=username)

    return HTMLResponse(str(page))


@router.post("/posts/{id}/edit")
async def handle_post_edit(
    id: int,
    username: dependencies.UsernameFromSessionCookie,
    bulletin_service: dependencies.BulletinService,
    title: str = Form(...),
    text: str = Form(...),
) -> Response:
    """Defines web endpoint for handling edit post."""
    if not username:
        raise http_exceptions.Forbidden

    await bulletin_service.update_post(
        id=id,
        username=username,
        payload=schemas.PostInputSchema(title=title, text=text),
    )

    return RedirectResponse(f"/posts/{id}", status_code=status.HTTP_302_FOUND)


@router.get("/posts/{id}/delete")
async def handle_post_delete_page(
    id: int,
    username: dependencies.UsernameFromSessionCookie,
    bulletin_service: dependencies.BulletinService,
) -> Response:
    """Defines web endpoint for displaying post to edit."""
    if not username:
        raise http_exceptions.Forbidden

    if (post := await bulletin_service.get_post(id=id)) is None:
        raise http_exceptions.NotFound

    content = H.main(local_components.post_form_delete(post))

    page = components.page(content, title="Post Delete", username=username)

    return HTMLResponse(str(page))


@router.post("/posts/{id}/delete")
async def handle_post_delete(
    id: int,
    username: dependencies.UsernameFromSessionCookie,
    bulletin_service: dependencies.BulletinService,
) -> Response:
    """Defines web endpoint for handling edit post."""
    if not username:
        raise http_exceptions.Forbidden

    await bulletin_service.delete_post(id=id)

    return RedirectResponse("/posts", status_code=status.HTTP_302_FOUND)



from pathlib import Path

from fastapi import FastAPI, status
from fastapi.exceptions import RequestValidationError
from fastapi.staticfiles import StaticFiles

import poaster.access.api
import poaster.access.web
import poaster.bulletin.api
import poaster.bulletin.web
from poaster.__about__ import __version__
from poaster.core import http_exceptions


def get_app() -> FastAPI:
    """Build and configure application server."""

    # main web application
    app = FastAPI(docs_url=None, redoc_url=None, openapi_url=None)

    # mount static files
    static_directory = Path(__file__).parent / "static"
    app.mount("/static", StaticFiles(directory=static_directory), name="static")

    # include web endpoints
    app.include_router(poaster.access.web.router)
    app.include_router(poaster.bulletin.web.router)

    # exception handling
    app.add_exception_handler(
        status.HTTP_403_FORBIDDEN, http_exceptions.handle_forbidden
    )
    app.add_exception_handler(
        status.HTTP_404_NOT_FOUND, http_exceptions.handle_not_found
    )
    app.add_exception_handler(
        RequestValidationError, http_exceptions.handle_validation_error
    )

    # json api
    api = FastAPI(
        title="poaster API",
        version=__version__,
        summary="Minimal, libre bulletin board for posts.",
        license_info={
            "name": "GNU Affero General Public License (AGPL)",
            "url": "https://www.gnu.org/licenses/agpl-3.0.html",
        },
    )

    # include api endpoints
    api.include_router(poaster.access.api.router)
    api.include_router(poaster.bulletin.api.router)

    # mount api app to web app
    app.mount("/api", api)

    return app


app = get_app()



class DoesNotExist(Exception):
    """Raise when entity not found in database."""


class AlreadyExists(Exception):
    """Raise when entity already exists."""


class Unauthorized(Exception):
    """Raise when authentication fails."""



import pytest
from click.testing import CliRunner
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession

from poaster.access import cli


@pytest.fixture
async def add_user(db_session: AsyncSession) -> None:
    qry = "INSERT OR IGNORE INTO users (username, password) VALUES ('testuser', 'password');"
    await db_session.execute(text(qry))


def test_users_help(cli_runner: CliRunner):
    expected_help_text = """\
Usage: users [OPTIONS] COMMAND [ARGS]...

  Control panel for managing users.

Options:
  -h, --help  Show this message and exit.

Commands:
  add   Add new user.
  list  List stored usernames.
"""
    result = cli_runner.invoke(cli.users, ["--help"])

    assert result.exit_code == 0
    assert result.output == expected_help_text


def test_users_add(cli_runner: CliRunner):
    add_user_args = ["add", "--username", "testuser", "--password", "password"]
    result = cli_runner.invoke(cli.users, add_user_args)

    assert result.exit_code == 0
    assert result.output == "`testuser` successfully added.\n"


def test_users_add_validation_failed(cli_runner: CliRunner):
    add_user_args = ["add", "--username", "testuser" * 100, "--password", "password"]
    result = cli_runner.invoke(cli.users, add_user_args)

    assert result.exit_code == 0
    assert "Input validation failed:" in result.output


@pytest.mark.usefixtures("add_user")
def test_users_add_already_exists(cli_runner: CliRunner):
    add_user_args = ["add", "--username", "testuser", "--password", "password"]
    result = cli_runner.invoke(cli.users, add_user_args)

    assert result.exit_code == 0
    assert result.output == "User already exists.\n"


@pytest.mark.usefixtures("add_user")
def test_users_list(cli_runner: CliRunner):
    result = cli_runner.invoke(cli.users, ["list"])

    assert result.exit_code == 0
    assert result.output == "Stored users:\n- testuser\n"


def test_users_list_no_users(cli_runner: CliRunner):
    result = cli_runner.invoke(cli.users, ["list"])

    assert result.exit_code == 0
    assert result.output == "No users found.\n"



import logging

import pytest
from uvicorn.logging import TRACE_LOG_LEVEL

from poaster.core.config import LogLevel, Settings, get_log_level, get_settings


def test_get_settings_from_env(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv("LOG_LEVEL", "critical")
    monkeypatch.setenv("DB_PATH", "/./test.db")
    monkeypatch.setenv("SECRET_KEY", "notsosecretive")
    monkeypatch.setenv("SECRET_KEY_N_BYTES", "42")
    monkeypatch.setenv("ALGORITHM", "secure")

    got = get_settings()
    want = Settings(
        log_level="critical",
        db_path="/./test.db",
        secret_key="notsosecretive",
        secret_key_n_bytes=42,
        algorithm="secure",
    )
    assert got == want


@pytest.mark.parametrize(
    "level, want",
    [
        ("critical", logging.CRITICAL),
        ("error", logging.ERROR),
        ("warning", logging.WARNING),
        ("info", logging.INFO),
        ("debug", logging.DEBUG),
        ("trace", TRACE_LOG_LEVEL),
    ],
)
def test_get_log_level(level: LogLevel, want: int):
    got = get_log_level(level)
    assert got == want


@pytest.mark.parametrize("level", ["fatal", "scary", "few"])
def test_get_bad_log_level(level: LogLevel):
    exc = f"{level=} invalid. Log levels: critical, error, warning, info, debug, trace"
    with pytest.raises(RuntimeError, match=exc):
        get_log_level(level)



import functools
import logging
from typing import Literal

from pydantic_settings import BaseSettings, SettingsConfigDict
from uvicorn.logging import TRACE_LOG_LEVEL

ColorVariant = Literal["danger", "info", "primary", "success", "warning"]
LogLevel = Literal["critical", "error", "warning", "info", "debug", "trace"]


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", extra="ignore", frozen=True)

    log_level: LogLevel = "info"
    db_path: str = "/./poaster.db"
    secret_key: str = ""
    secret_key_n_bytes: int = 32
    algorithm: str = "HS256"

    # theming
    title: str = "poaster"
    color_danger: str = "#FF595E"
    color_info: str = "#1982C4"
    color_primary: str = "#6A4C93"
    color_success: str = "#8AC926"
    color_warning: str = "#FFCA3A"

    @functools.cached_property
    def async_uri(self) -> str:
        return f"sqlite+aiosqlite://{self.db_path}"

    @functools.cached_property
    def uri(self) -> str:
        return f"sqlite://{self.db_path}"

    @functools.cached_property
    def _color_map(self) -> dict[ColorVariant, str]:
        return {
            "danger": self.color_danger,
            "info": self.color_info,
            "primary": self.color_primary,
            "success": self.color_success,
            "warning": self.color_warning,
        }

    def get_color(self, variant: ColorVariant) -> str:
        return self._color_map[variant]


def get_log_level(level: LogLevel) -> int:
    log_levels: dict[LogLevel, int] = {
        "critical": logging.CRITICAL,
        "error": logging.ERROR,
        "warning": logging.WARNING,
        "info": logging.INFO,
        "debug": logging.DEBUG,
        "trace": TRACE_LOG_LEVEL,
    }
    try:
        return log_levels[level]
    except KeyError:
        valid_levels = ", ".join(log_levels.keys())
        raise RuntimeError(f"{level=} invalid. Log levels: {valid_levels}") from None


def get_settings() -> Settings:
    """Load settings from environment and configure logging."""
    settings = Settings()

    logging.basicConfig(
        level=get_log_level(settings.log_level),
        format="%(asctime)s - %(levelname)s: %(message)s",
    )

    return settings


settings = get_settings()



import datetime

import pytest
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession

from poaster.bulletin import schemas, services


@pytest.fixture
def bulletin_service(db_session: AsyncSession) -> services.BulletinService:
    return services.BulletinService.from_session(db_session)


async def test_create_post(bulletin_service: services.BulletinService):
    got = await bulletin_service.create_post(
        username="testuser",
        payload=schemas.PostInputSchema(title="hi", text="hello, world!"),
    )
    want = schemas.PostSchema(
        id=1,
        title="hi",
        text="hello, world!",
        created_by="testuser",
        created_at=got.created_at,
    )

    assert got == want


async def test_update_post(
    db_session: AsyncSession, bulletin_service: services.BulletinService
):
    qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(qry))

    got = await bulletin_service.update_post(
        id=1,
        username="testuser",
        payload=schemas.PostInputSchema(title="hi v2", text="hello, world!!"),
    )
    want = schemas.PostSchema(
        id=1,
        title="hi v2",
        text="hello, world!!",
        created_by="testuser",
        created_at=datetime.datetime(2024, 2, 13, 8),
    )

    assert got == want


async def test_update_post_not_found(bulletin_service: services.BulletinService):
    got = await bulletin_service.update_post(
        id=42,
        username="testuser",
        payload=schemas.PostInputSchema(title="hi", text="hello, world!"),
    )
    want = None

    assert got == want


async def test_delete_post(
    db_session: AsyncSession, bulletin_service: services.BulletinService
):
    qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(qry))

    got = await bulletin_service.delete_post(id=1)
    want = schemas.PostSchema(
        id=1,
        title="hi",
        text="hello, world!",
        created_by="testuser",
        created_at=datetime.datetime(2024, 2, 13, 8),
    )

    assert got == want


async def test_delete_post_not_found(bulletin_service: services.BulletinService):
    got = await bulletin_service.delete_post(id=42)
    want = None
    assert got == want


async def test_get_post(
    db_session: AsyncSession, bulletin_service: services.BulletinService
):
    qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(qry))

    got = await bulletin_service.get_post(id=1)
    want = schemas.PostSchema(
        id=1,
        title="hi",
        text="hello, world!",
        created_by="testuser",
        created_at=datetime.datetime(2024, 2, 13, 8),
    )

    assert got == want


async def test_get_post_not_found(bulletin_service: services.BulletinService):
    got = await bulletin_service.get_post(id=42)
    want = None

    assert got == want


async def test_get_posts(
    db_session: AsyncSession, bulletin_service: services.BulletinService
):
    qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES
        ('oldest_post', 'hello, oldest_post!', 'testuser', '2024-02-13 08:00'),
        ('newest_post', 'hello, newest_post!', 'testuser', '2024-02-15 12:00')
    ;
    """
    await db_session.execute(text(qry))

    got = await bulletin_service.get_posts()
    want = [
        schemas.PostSchema(
            id=2,
            title="newest_post",
            text="hello, newest_post!",
            created_by="testuser",
            created_at=datetime.datetime(2024, 2, 15, 12, 0),
        ),
        schemas.PostSchema(
            id=1,
            title="oldest_post",
            text="hello, oldest_post!",
            created_by="testuser",
            created_at=datetime.datetime(2024, 2, 13, 8, 0),
        ),
    ]

    assert got == want


async def test_get_all_none_found(bulletin_service: services.BulletinService):
    got = await bulletin_service.get_posts()
    want = []

    assert got == want


async def test_get_latest_version_of_post(
    db_session: AsyncSession, bulletin_service: services.BulletinService
):
    add_post_qry = """
    INSERT INTO posts (title, text, created_by, created_at)
    VALUES ('hi', 'hello, world!', 'testuser', '2024-02-13 08:00');
    """
    await db_session.execute(text(add_post_qry))

    add_post_versions_qry = """
    INSERT INTO post_versions (post_id, title, text, version, updated_by, updated_at)
    VALUES
      (1, 'hi v1', 'hello, world!', 1, 'testuser', '2024-02-13 08:00'),
      (1, 'hi v2', 'hello, world!!', 2, 'testuser', '2024-02-13 09:00'),
      (1, 'hi v3', 'hello, world!!!', 3, 'testuser', '2024-02-13 10:00')
    ;
    """
    await db_session.execute(text(add_post_versions_qry))

    got = await bulletin_service.get_latest_version_of_post(id=1)
    want = schemas.PostVersionSchema(
        id=3,
        post_id=1,
        title="hi v3",
        text="hello, world!!!",
        version=3,
        updated_by="testuser",
        updated_at=datetime.datetime(2024, 2, 13, 10),
    )

    assert got == want



import pytest
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession

from poaster.access import repository, schemas
from poaster.core import exceptions


@pytest.fixture
def user_repo(db_session: AsyncSession) -> repository.SqlalchemyUserRepository:
    return repository.SqlalchemyUserRepository(db_session)


async def test_create_user_hashes_password(
    user_repo: repository.SupportsUserRepository,
):
    user = schemas.UserRegistrationSchema(username="bob", password="password")
    db_user = await user_repo.create(user)
    assert db_user.username == "bob"
    assert db_user.password.startswith("$argon2")


async def test_can_create_user(user_repo: repository.SupportsUserRepository):
    user = schemas.UserRegistrationSchema(username="bob", password="password")

    got = await user_repo.create(user)
    want = schemas.UserSchema(
        id=got.id,
        username="bob",
        password=got.password,  # dynamically hashed password
    )

    assert got == want


async def test_duplicate_user_raises(user_repo: repository.SupportsUserRepository):
    user = schemas.UserRegistrationSchema(username="bob", password="password")
    await user_repo.create(user)
    with pytest.raises(exceptions.AlreadyExists):
        await user_repo.create(user)


async def test_user_is_found_by_username(
    db_session: AsyncSession, user_repo: repository.SupportsUserRepository
):
    qry = "INSERT INTO users (username, password) VALUES ('bob', 'hashedpw');"
    await db_session.execute(text(qry))

    got = await user_repo.get_by_username("bob")
    want = schemas.UserSchema(
        id=got.id,
        username="bob",
        password="hashedpw",
    )

    assert got == want


async def test_user_is_not_found_by_username(
    user_repo: repository.SupportsUserRepository,
):
    with pytest.raises(exceptions.DoesNotExist):
        await user_repo.get_by_username("nobody")


async def test_get_all_users(
    db_session: AsyncSession, user_repo: repository.SupportsUserRepository
):
    qry = "INSERT INTO users (username, password) VALUES ('bob', 'hashedpw');"
    await db_session.execute(text(qry))

    got = await user_repo.get_all()
    want = [schemas.UserSchema(id=1, username="bob", password="hashedpw")]

    assert got == want


async def test_get_all_users_none_found(user_repo: repository.SupportsUserRepository):
    got = await user_repo.get_all()
    want = []

    assert got == want



from typing import Protocol

from sqlalchemy import delete, select, update
from sqlalchemy.ext.asyncio import AsyncSession

from poaster.core import exceptions

from . import schemas, tables


class SupportsPostRepository(Protocol):
    """Interface for handling posts."""

    async def create(
        self, username: str, post: schemas.PostInputSchema
    ) -> schemas.PostSchema:
        """Create post after validating input schema."""
        ...

    async def update(
        self, id: int, post: schemas.PostInputSchema
    ) -> schemas.PostSchema:
        """Update post after validating input schema."""
        ...

    async def delete(self, id: int) -> schemas.PostSchema:
        """Delete post by id."""
        ...

    async def get_all(self) -> list[schemas.PostSchema]:
        """Fetch all posts from the DB."""
        ...

    async def get_by_id(self, id: int) -> schemas.PostSchema:
        """Fetch by id, raising exception if not found."""
        ...


class SqlalchemyPostRepository:
    """Implementation of the post repository with SqlAlchemy."""

    def __init__(self, session: AsyncSession) -> None:
        self._session = session

    async def create(
        self, username: str, post: schemas.PostInputSchema
    ) -> schemas.PostSchema:
        entry = tables.Post(**post.model_dump(), created_by=username)
        self._session.add(entry)
        await self._session.commit()
        return schemas.PostSchema.model_validate(entry)

    async def update(
        self, id: int, post: schemas.PostInputSchema
    ) -> schemas.PostSchema:
        if await self._session.get(tables.Post, id) is None:
            raise exceptions.DoesNotExist("Post doesn't exist.")

        qry = (
            update(tables.Post)
            .where(tables.Post.id == id)
            .values(title=post.title, text=post.text)
        )
        await self._session.execute(qry)
        await self._session.commit()

        db_post = await self._session.get(tables.Post, id)

        return schemas.PostSchema.model_validate(db_post)

    async def delete(self, id: int) -> schemas.PostSchema:
        if (post := await self._session.get(tables.Post, id)) is None:
            raise exceptions.DoesNotExist("Post doesn't exist.")

        qry = delete(tables.Post).where(tables.Post.id == post.id)
        await self._session.execute(qry)
        await self._session.commit()

        return schemas.PostSchema.model_validate(post)

    async def get_all(self) -> list[schemas.PostSchema]:
        qry = select(tables.Post).order_by(tables.Post.created_at.desc())
        results = await self._session.execute(qry)
        return [schemas.PostSchema.model_validate(res) for res in results.scalars()]

    async def get_by_id(self, id: int) -> schemas.PostSchema:
        if (post := await self._session.get(tables.Post, id)) is None:
            raise exceptions.DoesNotExist("Post doesn't exist.")

        return schemas.PostSchema.model_validate(post)


class SupportsPostVersionRepository(Protocol):
    """Interface for handling post versions."""

    async def create(
        self, *, username: str, post_id: int, post: schemas.PostInputSchema
    ) -> schemas.PostVersionSchema:
        """Create post version after validating input schema."""
        ...

    async def get_all(self, *, post_id: int) -> list[schemas.PostVersionSchema]:
        """Fetch all versions of a post from the DB."""
        ...

    async def get_latest(self, *, post_id: int) -> schemas.PostVersionSchema:
        """Fetch the latest version of a post from the DB."""
        ...

    async def get_by_id(
        self, *, post_id: int, version_id: int
    ) -> schemas.PostVersionSchema:
        """Fetch post version by id, raising exception if not found."""
        ...


class SqlalchemyPostVersionRepository:
    """Implementation of the post version repository with SqlAlchemy."""

    def __init__(self, session: AsyncSession) -> None:
        self._session = session

    async def create(
        self, *, username: str, post_id: int, post: schemas.PostInputSchema
    ) -> schemas.PostVersionSchema:
        qry = (
            select(tables.PostVersion)
            .where(tables.PostVersion.post_id == post_id)
            .order_by(tables.PostVersion.updated_at.desc())
        )
        result = await self._session.execute(qry)
        latest_version = result.scalars().first()

        next_version = (
            schemas.PostVersionSchema.model_validate(latest_version).version + 1
            if latest_version
            else 1
        )

        entry = tables.PostVersion(
            **post.model_dump(),
            post_id=post_id,
            updated_by=username,
            version=next_version,
        )

        self._session.add(entry)
        await self._session.commit()
        return schemas.PostVersionSchema.model_validate(entry)

    async def get_all(self, *, post_id: int) -> list[schemas.PostVersionSchema]:
        qry = (
            select(tables.PostVersion)
            .where(tables.PostVersion.post_id == post_id)
            .order_by(tables.PostVersion.updated_at)
        )
        results = await self._session.execute(qry)
        return [
            schemas.PostVersionSchema.model_validate(post_version)
            for post_version in results.scalars()
        ]

    async def get_latest(self, *, post_id: int) -> schemas.PostVersionSchema:
        qry = (
            select(tables.PostVersion)
            .where(tables.PostVersion.post_id == post_id)
            .order_by(tables.PostVersion.updated_at.desc())
        )
        results = await self._session.execute(qry)
        latest_post_version = results.scalars().first()
        return schemas.PostVersionSchema.model_validate(latest_post_version)

    async def get_by_id(
        self, *, post_id: int, version_id: int
    ) -> schemas.PostVersionSchema:
        qry = (
            select(tables.PostVersion)
            .where(tables.PostVersion.post_id == post_id)
            .where(tables.PostVersion.version == version_id)
        )
        result = await self._session.execute(qry)

        if (post_version := result.scalars().one_or_none()) is None:
            raise exceptions.DoesNotExist("Post version doesn't exist.")

        return schemas.PostVersionSchema.model_validate(post_version)
